﻿-- Synchronizing module

-- spinning tiles fading in/out times
local SPINNING_TILE_FADE_IN_TIME = 0.07
local SPINNING_TILE_FADE_OUT_TIME = 0.37
local SPINNING_TILE_SIZE = 27

-- client syncing state
-- careful about changing 0, it is hardcoded in the main module
local SYNCING_STATE_NONE = 0
-- set if waiting for newer versions (upon logging in/getting into party or raid)
local SYNCING_STATE_CHECK_VERSION = 1
-- set if this has performed a new search
local SYNCING_STATE_NEW_SEARCH = 2
-- set if received new search key and started listening
local SYNCING_STATE_RECEIVING_NEW_SEARCH = 3
-- set if sender sent all the items (after own search or logging in) and is waiting for more accurate predictions
local SYNCING_STATE_NARROWING_SEARCH = 4
-- set if receiver is waiting for full item details (after receiving short item details, recalculating new predictions and sending them back)
local SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS = 5
-- set if receiver has just logged in and is waiting for newer searches or better predictions
local SYNCING_STATE_AWAITING_BETTER_SEARCH = 6
BMTSyncingState = SYNCING_STATE_NONE

-- syncing state texts (for spinning inset in the main frame)
local SYNCING_STATE_CHECK_VERSION_TEXT = "Synchronizing Version"
local SYNCING_STATE_NEW_SEARCH_TEXT = "Synchronizing Items - New Search"
local SYNCING_STATE_NARROWING_SEARCH_TEXT = "Attempting to narrow time predictions"
local SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS_TEXT = "Waiting for full item details"
local SYNCING_STATE_AWAITING_BETTER_SEARCH_TEXT = "Checking for more recent search"

-- timeouts (in seconds)
-- after 2 seconds further syncing steps will be taken
local SYNCING_TIMEOUT_VERSION = 5
-- time to send out all the items after starting to send them, otherwise everything will be ignored
local SYNCING_TIMEOUT_NEW_SEARCH = 5
-- same as above, but when requested upon logging in etc.
local SYNCING_TIMEOUT_BETTER_SEARCH_REQ = 5
-- full item details timeout (if waiting for those)
local SYNCING_TIMEOUT_FULL_ITEM_DETAILS = 8

-- addon message prefix
local ADDON_PREFIX = "BMT"
-- addon message separator (fields)
local SEPARATOR = "%"
-- addon message separator (tables/category)
local SEPARATOR_CATEGORY = "$"
-- syncing keys (strings) sent through addon message
-- version request (also includes this version)
local KEY_VERREQ = "VERREQ"
-- version response
local KEY_VERRESP = "VERRESP"
-- new search key
local KEY_NEW_SEARCH = "NSEA"
-- better predictions resync key
local KEY_BETTER_PREDICTIONS = "BPRED"
-- better search request
local KEY_BETTER_SEARCH_REQ = "BSEAREQ"
-- table for easier parsing (update info)
local updateInfoSyncField = {"playerName",	"playerClass",	"playerGuild",	"updateDate",	"updateTime" }
local isUpdateFieldNumber = {false,			false,			false,			false,			true }
-- short item details key (used for deciding wheather receiver's auction should be moved to history, as well as determining new predictions)
local KEY_ITEM_SHORT = "ISHO"
-- table for easier parsing (update info)
local itemShortSyncField = {"name", "marketID", "timeLeft", "currBid", "numBids", "lastTimeLeftChangeDate", "lastTimeLeftChangeTime", "predictedLow", "predictedHigh" }
local isItemShortFieldNumber = {false, true, true, true, true, false, true, true, true}
-- full item details
local KEY_ITEM_FULL = "IFULL"
local itemFullSyncField = {"name", "marketID", "timeLeft", "currBid", "numBids", "lastTimeLeftChangeDate", "lastTimeLeftChangeTime", "predictedLow", "predictedHigh", "level", "highestBidder", "texture", "itemType", "link", "sellerName", "minIncrement", "quantity", "levelType", "minBid", "predictionsUpdateDiff"}
local isItemFullSyncFieldNumber = {false, true, true, true, true, false, true, true, true, true, false, false, false, false, false, true, true, false, true, true}

local ITEMS_NONE = "NO_ITEMS"

-- true if addon message broadcast is queued, false otherwise (done to avoid spam of system messages
-- about player being offline)
BMTBroadcastQueued = false
-- variable contaiting string to be broadcasted
local messageToBroadcast = nil;


-- highest revision received from others upon logging in
BMTHighestRevision = BMTRevision;

-- this is a 'copy' of current db for realm received from sync
-- the fields' name should be the same as in the original db
BMTSyncDB = {}

-- this is also a 'copy' of current db for received realm, but 
-- it can store multiple versions of it for the same realm
BMTSyncBetterDB = {}

--[[
function BMTSyncing_Initialize(configDB)
	config = BMT.db.profile
end
]]

function BMTSyncing_SendAddonMessage(msg, channel, target)
	if (channel == "GUILD" and BMTConfig.syncInGuild and IsInGuild()) or (channel == "WHISPER" and BMTConfig.syncWithFriends) then
		BMT:SendCommMessage(ADDON_PREFIX, msg, channel, target)
	end
end

-- this function doesn't actually perform a message broadcast
function BMTSyncing_BroadcastAddonMessage(msg)
	messageToBroadcast = msg
	BMTBroadcastQueued = true
	ShowFriends()
	-- callback done in main module that will catch the friends upate event
end

-- called ONLY from main module
function BMTSyncing_PerformRealBroadcast()
	if BMTBroadcastQueued then 
		BMTBroadcastQueued = false
	end
	if messageToBroadcast == nil then return end
	-- guild
	BMTSyncing_SendAddonMessage(messageToBroadcast, "GUILD")
	-- non-BNET friends
	local numAddonFriends = #BMTConfig.syncedFriends
	local numGameFriends = GetNumFriends()
	for i=1, numAddonFriends do
		local name = BMTConfig.syncedFriends[i]
		-- find in in-game list and check if is online
		for j=1, numGameFriends do
			local n, _, _, _, connected = GetFriendInfo(j)
			if name == n then
				if connected then
					BMTSyncing_SendAddonMessage(messageToBroadcast, "WHISPER", name)
				end
			end
		end
	end
	-- TODO: party/raid (check if in one)
	-- TODO: BNET
	messageToBroadcast = nil
end

function BMT:OnCommReceived(prefix, msg, channel, sender)
	if string.len(msg) > 70 then
		BMT:Debug("("..sender..")"..string.sub(msg, 1, 70).."...")
	else
		BMT:Debug("("..sender..")"..msg)
	end
	
	if sender == UnitName("player") then return end
	-- categories (initial info/update info/short items/full items etc.)
	-- initial - sender's addon version and realm being synced
	local categories = {strsplit(SEPARATOR_CATEGORY, msg)}
	if categories == nil then return end
	local initial = {strsplit(SEPARATOR, categories[1])}
	if initial == nil then return end
	local key = initial[1]
	local senderAddonVer = tonumber(initial[2])
	local syncedRealm = initial[3]
	if key == KEY_VERREQ then
		if senderAddonVer > BMTHighestRevision then
			BMTHighestRevision = senderAddonVer
		end
		-- answer using the same channel
		BMTSyncing_SendAddonMessage(BMTSyncing_InitialInfo(KEY_VERRESP), channel, sender)
	-- ignore messages other than version sync from channels that this isn't syncing with
	-- TODO: BNET
	elseif (channel == "GUILD" and BMTConfig.syncInGuild) or (channel == "WHISPER" and BMTConfig.syncWithFriends) then
		if key == KEY_VERRESP then
			if senderAddonVer > BMTHighestRevision then
				BMTHighestRevision = senderAddonVer
			end
		elseif key == KEY_NEW_SEARCH and BMTSyncingState == SYNCING_STATE_NONE then
			BMTSyncing_OnNewSearchReceive(categories, syncedRealm, channel, sender)
		elseif key == KEY_BETTER_PREDICTIONS  and (BMTSyncingState == SYNCING_STATE_NEW_SEARCH or BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH) then
			BMTSyncing_OnBetterResultReceive(categories, syncedRealm, channel, sender)
		-- always respond to this, regardless of syncing state
		elseif key == KEY_BETTER_SEARCH_REQ and BMTSyncingState == SYNCING_STATE_NONE then
			BMTSyncng_OnSearchRequestReceive(categories, syncedRealm, channel, sender)
		-- we also write the items if we're doing nothing (for some reason..)
		elseif key == KEY_ITEM_FULL and BMTSyncingState == SYNCING_STATE_NONE or BMTSyncingState == SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS then
			BMTSyncng_OnFullItemDetailsReceive(categories, syncedRealm, channel, sender)
		end
	end
end

-- categories[1] is initial update, 2 is updateInfo, rest is items
function BMTSyncing_OnNewSearchReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm].updateInfo = {}
	BMTSyncDB[syncedRealm].items = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	
	-- ignore searches older than our
	local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm].updateInfo.updateDate, BMTSyncDB[syncedRealm].updateInfo.updateTime, BMTGlobal.current[syncedRealm].updateInfo.updateDate, BMTGlobal.current[syncedRealm].updateInfo.updateTime)
	if res == "smaller" or res == "equal" then BMTSyncDB[syncedRealm] = nil return end
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncDB[syncedRealm].items
		-- (i-2) so we start from 1
		BMTSyncDB[syncedRealm].items[(i-2)] = {}
		for j=1, #itemShortSyncField do
			if isItemShortFieldNumber[j] then
				BMTSyncDB[syncedRealm].items[(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncDB[syncedRealm].items[(i-2)][itemShortSyncField[j]] = itemInfo[j]
			end
		end
	end
	
	-- rewrite update info to current
	BMTGlobal.current[syncedRealm].updateInfo = BMTSyncDB[syncedRealm].updateInfo
	
	-- here we have everything that will be needed to recalculate better predictions (and if needed, sending them back)
	BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncDB[syncedRealm].items, BMTSyncDB[syncedRealm].updateInfo.updateDate, BMTSyncDB[syncedRealm].updateInfo.updateTime)
	
	-- here, we only have items that should be on the current list in the BMTGlobal.current
	-- time to do predictions on our side (treat the search as if it was our own and check if they are better than the received ones)
	local betterPredMsg = BMTSyncing_InitialInfo(KEY_BETTER_PREDICTIONS)
	-- add update info (which is the same as the sender, but we want this message to look the same for both: after search narrowing and after login narrowing)
	for i=1, #updateInfoSyncField do
		if i > 1 then
			betterPredMsg = betterPredMsg .. SEPARATOR
		end
		betterPredMsg = betterPredMsg..BMTGlobal.current[syncedRealm].updateInfo[updateInfoSyncField[i]]
	end
	betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
	local numItems = BMT_GetNumItems(syncedRealm)
	local currDate = date("%d/%m/%y")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	for i=1, numItems do
		for j=1, #BMTSyncDB[syncedRealm].items do
			if BMTGlobal.current[syncedRealm].items[i].marketID == BMTSyncDB[syncedRealm].items[j].marketID then
				local newTimeLeft = BMTSyncDB[syncedRealm].items[j].timeLeft
				-- recalculate predictions
				BMT_RecalculateAndSetPredictedTimes(syncedRealm, i, newTimeLeft, currDate, currTime, false)
				-- check if better
				if BMTGlobal.current[syncedRealm].items[i].predictedLow > BMTSyncDB[syncedRealm].items[j].predictedLow or BMTGlobal.current[syncedRealm].items[i].predictedHigh < BMTSyncDB[syncedRealm].items[j].predictedHigh then
					-- concatenate this item to the message that will be sent back
					for k=1, #itemShortSyncField do
						if k > 1 then
							betterPredMsg = betterPredMsg .. SEPARATOR
						end
						betterPredMsg = betterPredMsg..BMTGlobal.current[syncedRealm].items[i][itemShortSyncField[k]]
					end
					betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
				end
				break
			end
		end
	end
	-- cut out any unnecessary separators
	if string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR_CATEGORY or string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR then
		betterPredMsg = string.sub(betterPredMsg, 1, string.len(betterPredMsg)-1)
	end
	-- send it back
	BMTSyncing_SendAddonMessage(betterPredMsg, channel, sender)
	-- clean up and start waiting for a the best result after sender puts everything together
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS)
	-- schedule the timeout function that sets the state back to none, in case someones' addon crashed/one logged our right between the messages etc.
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_FULL_ITEM_DETAILS, BMTSyncing_IgnoreFullItemDetails)
end

-- timeout function, just a cleanup so that we don't wait for the item details forever
function BMTSyncing_IgnoreFullItemDetails()
	BMT:Debug("Full item details timeout")
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
end

-- this function just adds received informations
function BMTSyncing_OnBetterResultReceive(categories, syncedRealm, channel, sender)
	BMT:Debug("Received better predictions from "..sender.." syncedRealm="..syncedRealm.." channel="..channel)
	if BMTSyncBetterDB[syncedRealm] == nil then
		BMTSyncBetterDB[syncedRealm] = {}
	end
	local index = #BMTSyncBetterDB[syncedRealm]+1
	BMTSyncBetterDB[syncedRealm][index] = {}
	BMTSyncBetterDB[syncedRealm][index].updateInfo = {}
	BMTSyncBetterDB[syncedRealm][index].items = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	-- DEBUG
	for i=1, #updateInfo do
		BMT:Debug("updateInfo["..i.."]="..updateInfo[i])
	end
	-- END OF DEBUG
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncBetterDB[syncedRealm][index].updateInfo[updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncBetterDB[syncedRealm][index].updateInfo[updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	
	--[[ not necessarily true, as we still want the DB that has more accurate predictions, even if it's older than our
	-- ignore searches older than our
	local res = BMT_CompareDateAndTime(BMTSyncBetterDB[syncedRealm][index].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][index].updateInfo.updateTime, BMTGlobal.current[syncedRealm].updateInfo.updateDate, BMTGlobal.current[syncedRealm].updateInfo.updateTime)
	if res == "smaller" or res == "equal" then BMTSyncBetterDB[syncedRealm][index] = nil return end
	]]
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncBetterDB[syncedRealm][index].items
		-- (i-2) so we start from 1
		BMTSyncBetterDB[syncedRealm][index].items[(i-2)] = {}
		for j=1, #itemShortSyncField do
			if isItemShortFieldNumber[j] then
				BMTSyncBetterDB[syncedRealm][index].items[(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncBetterDB[syncedRealm][index].items[(i-2)][itemShortSyncField[j]] = itemInfo[j]
			end
			if itemShortSyncField[j] == "predictedLow" and BMTSyncBetterDB[syncedRealm][index].items[(i-2)][itemShortSyncField[j]] < 0 then
				BMTSyncBetterDB[syncedRealm][index].items[(i-2)][itemShortSyncField[j]] = 0
			end
		end
	end
	BMT:Debug("Received better predictions:")
	BMT:Debug("numUpdateFields="..#BMTSyncBetterDB[syncedRealm][index].updateInfo)
	BMT:Debug("numItems="..#BMTSyncBetterDB[syncedRealm][index].items)
end


function BMTSyncng_OnSearchRequestReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm].updateInfo = {}
	BMTSyncDB[syncedRealm].items = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	local haveNewerSearch = false
	local haveBetterItems = false
	local incomingDBEmpty = false
	local ourDBEmpty = false
	-- check for empty incoming db
	if updateInfo[1] == ITEMS_NONE then
		incomingDBEmpty = true
		BMT:Debug("Incoming better search request from "..sender.." has empty db")
	end
	local betterPredMsg = BMTSyncing_InitialInfo(KEY_BETTER_PREDICTIONS)
	-- check for our empty db
	if BMTGlobal.current[syncedRealm].updateInfo.playerName == nil then
		ourDBEmpty = true
	end
	if incomingDBEmpty == false then
		for i=1, #updateInfoSyncField do
			if isUpdateFieldNumber[i] then
				BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = tonumber(updateInfo[i])
			else
				BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = updateInfo[i]
			end
		end
		
		if ourDBEmpty == false then
			-- we only let one know if we have newer search or more accurate predictions OR newer search
			local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm].updateInfo.updateDate, BMTSyncDB[syncedRealm].updateInfo.updateTime, BMTGlobal.current[syncedRealm].updateInfo.updateDate, BMTGlobal.current[syncedRealm].updateInfo.updateTime)
			if res == "greater"  then 
				haveNewerSearch = true
			end
		end
		
		-- go through all of the items (items start @ index 3)
		for i=3, #categories do
			local itemInfo = {strsplit(SEPARATOR, categories[i])}
			local numItems = #BMTSyncDB[syncedRealm].items
			-- (i-2) so we start from 1
			BMTSyncDB[syncedRealm].items[(i-2)] = {}
			for j=1, #itemShortSyncField do
				if isItemShortFieldNumber[j] then
					BMTSyncDB[syncedRealm].items[(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
				else
					BMTSyncDB[syncedRealm].items[(i-2)][itemShortSyncField[j]] = itemInfo[j]
				end
			end
		end
	end

	
	-- if our DB was empty and incoming was not, let's set the incoming info to current and send it back already (so the message format is the same across all cases)
	if ourDBEmpty == true and incomingDBEmpty == false then
		BMTGlobal.current[syncedRealm].updateInfo = BMTSyncDB[syncedRealm].updateInfo
	end
	-- add update info to response message
	for i=1, #updateInfoSyncField do
		if i > 1 then
			betterPredMsg = betterPredMsg .. SEPARATOR
		end
		betterPredMsg = betterPredMsg..BMTGlobal.current[syncedRealm].updateInfo[updateInfoSyncField[i]]
	end
	betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
	local numItems = BMT_GetNumItems(syncedRealm)
	local currDate = date("%d/%m/%y")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	-- refresh all predictions before checking anything
	BMT_RefreshAllPredictedTimes(syncedRealm)
	-- this case compares incoming and our own db (therefore both have to exist) and sends back only better items
	if incomingDBEmpty == false then
		if ourDBEmpty == false then
			for i=1, numItems do
				for j=1, #BMTSyncDB[syncedRealm].items do
					if BMTGlobal.current[syncedRealm].items[i].marketID == BMTSyncDB[syncedRealm].items[j].marketID then
						-- check if better
						if BMTGlobal.current[syncedRealm].items[i].predictedLow > BMTSyncDB[syncedRealm].items[j].predictedLow or BMTGlobal.current[syncedRealm].items[i].predictedHigh < BMTSyncDB[syncedRealm].items[j].predictedHigh then
							haveBetterItems = true
							-- concatenate this item to the message that will be sent back
							for k=1, #itemShortSyncField do
								if k > 1 then
									betterPredMsg = betterPredMsg .. SEPARATOR
								end
								betterPredMsg = betterPredMsg..BMTGlobal.current[syncedRealm].items[i][itemShortSyncField[k]]
							end
							betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
						end
						break
					end
				end
			end
		end
	-- no incoming items, send everything (if we have anything)
	else
		if ourDBEmpty == false then
			haveBetterItems = true
			for i=1, numItems do
				for k=1, #itemShortSyncField do
					if k > 1 then
						betterPredMsg = betterPredMsg .. SEPARATOR
					end
					betterPredMsg = betterPredMsg..BMTGlobal.current[syncedRealm].items[i][itemShortSyncField[k]]
				end
				betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
			end
		end
	end
	-- cut out any unnecessary separators
	while string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR_CATEGORY or string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR do
		betterPredMsg = string.sub(betterPredMsg, 1, string.len(betterPredMsg)-1)
	end
	-- send it back
	if haveNewerSearch == true or haveBetterItems == true or incomingDBEmpty == true then
		-- respond
		BMTSyncing_SendAddonMessage(betterPredMsg, channel, sender)
	end
	-- clean up and start waiting for a the best result after sender puts everything together
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_FULL_ITEM_DETAILS, BMTSyncing_IgnoreFullItemDetails)
end

-- this should contain the newest and most accurate search
-- if the search is older than our - return
-- if any item is less accurate than our, don't write it
function BMTSyncng_OnFullItemDetailsReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm].updateInfo = {}
	BMTSyncDB[syncedRealm].items = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncDB[syncedRealm].updateInfo[updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm].updateInfo.updateDate, BMTSyncDB[syncedRealm].updateInfo.updateTime, BMTGlobal.current[syncedRealm].updateInfo.updateDate, BMTGlobal.current[syncedRealm].updateInfo.updateTime)
	if res == "smaller" then
		-- clean up
		BMTSyncDB[syncedRealm] = nil
		BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
		return 
	end
	
	BMT:Debug("debug 1: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncDB[syncedRealm].items
		-- (i-2) so we start from 1
		BMTSyncDB[syncedRealm].items[(i-2)] = {}
		for j=1, #itemFullSyncField do
			if isItemFullSyncFieldNumber[j] then
				BMTSyncDB[syncedRealm].items[(i-2)][itemFullSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncDB[syncedRealm].items[(i-2)][itemFullSyncField[j]] = itemInfo[j]
			end
		end
	end
	
	BMT:Debug("recv: "..BMTSyncDB[syncedRealm].items[1].predictedLow)
	
	BMT:Debug("debug 2: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	
	-- rewrite update info to current
	BMTGlobal.current[syncedRealm].updateInfo = BMTSyncDB[syncedRealm].updateInfo
	BMT:Debug("debug 3: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	-- do this just in case
	BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncDB[syncedRealm].items, BMTSyncDB[syncedRealm].updateInfo.updateDate, BMTSyncDB[syncedRealm].updateInfo.updateTime)
	BMT:Debug("debug 4: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	-- refresh predictions before doing anything
	BMT_RefreshAllPredictedTimes(syncedRealm)
	BMT:Debug("debug 5: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	BMT_AddOrUpdateExisting(syncedRealm, BMTSyncDB[syncedRealm].items, true)
	--[[
	local numItems = BMT_GetNumItems(syncedRealm)
	for i=1, numItems do
		for j=1, #BMTSyncDB[syncedRealm].items do
			if BMTGlobal.current[syncedRealm].items[i].marketID == BMTSyncDB[syncedRealm].items[j].marketID then
				-- check if better (note: this is the other way around than in the rest of the code)
				if BMTGlobal.current[syncedRealm].items[i].predictedLow < BMTSyncDB[syncedRealm].items[j].predictedLow or BMTGlobal.current[syncedRealm].items[i].predictedHigh > BMTSyncDB[syncedRealm].items[j].predictedHigh then
					-- rewrite everything to current
					BMTGlobal.current[syncedRealm].items[i] = BMTSyncDB[syncedRealm].items[j]
				end
				break
			end
		end
	end
	]]
	BMT:Debug("debug 6: "..BMTGlobal.current[syncedRealm].items[1].predictedLow)
	-- clean up
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
	
	-- for bottom text handling only
	BMT_SetNewUpdater(syncedRealm, BMTGlobal.current[syncedRealm].updateInfo.playerName, BMTGlobal.current[syncedRealm].updateInfo.playerClass, BMTGlobal.current[syncedRealm].updateInfo.playerGuild, BMTGlobal.current[syncedRealm].updateInfo.updateTime, BMTGlobal.current[syncedRealm].updateInfo.updateDate);
	-- update the list in case the addon's browser is opened
	-- bottom text is already updated in BMT_SetNewUpdater()
	if(BMTFrame:IsShown()) then
		BMTScrollFrame_Update();
	end
	BMTTaskManager_Unschedule(BMTSyncing_IgnoreFullItemDetails)
end

-- mainly to handle the spinning inset
function BMTSyncing_ChangeSyncingState(newState)
	BMTSyncingState = newState
	if BMTFrame:IsShown() then
		if BMTSyncingState == 0 then
			BMTFrameSpinningInset:Hide()
		else
			BMTFrameSpinningInset:Show()
		end
	end
	if BMTSyncingState == SYNCING_STATE_CHECK_VERSION then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_CHECK_VERSION_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_NEW_SEARCH or BMTSyncingState == SYNCING_STATE_RECEIVING_NEW_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_NEW_SEARCH_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_NARROWING_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_NARROWING_SEARCH_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_AWAITING_BETTER_SEARCH_TEXT)
	end
end

-- returns initial informations about sender that is added to every addon message
-- parameters:
--	- key - key of the sync
-- returns: string in format key%addonVer%realm$
-- where	% is a separator within a group
-- 			$ is a category separator
function BMTSyncing_InitialInfo(key)
	local realmName = GetRealmName()
	local msg = key..SEPARATOR..tostring(BMTRevision)..SEPARATOR..realmName..SEPARATOR_CATEGORY
	
	return msg
end

-- asks for new version everywhere (called only upon logging in)
function BMTSyncing_AskForVersion_OnLogin()
	BMT:RegisterComm(ADDON_PREFIX, "OnCommReceived")
	
	-- no need to check if window is shown here, as this only happens upon logging in -> will be hidden
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_CHECK_VERSION)
	BMTSyncing_BroadcastAddonMessage(KEY_VERREQ..SEPARATOR..BMTRevision)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_VERSION, BMTSyncing_FinishSyncingVersion)
end

function BMTSyncing_BroadcastCurrentSearch()
	local realmName = GetRealmName()
	local msg = BMTSyncing_InitialInfo(KEY_NEW_SEARCH)
	-- last one has different separator
	for i=1, #updateInfoSyncField do
		if i > 1 then
			msg = msg .. SEPARATOR
		end
		msg = msg..BMTGlobal.current[realmName].updateInfo[updateInfoSyncField[i]]
	end
	msg = msg .. SEPARATOR_CATEGORY
	-- send out short details about each item, so that players can determine wheather some of their items are outdated or not, as well as determine new time predictions for each
	local numItems = BMT_GetNumItems(realmName)
	for j=1, numItems do
		for i=1, #itemShortSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BMTGlobal.current[realmName].items[j][itemShortSyncField[i]]
		end
		if j < numItems then
			msg = msg .. SEPARATOR_CATEGORY
		end
	end
	BMTSyncing_BroadcastAddonMessage(msg)
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NARROWING_SEARCH)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_NEW_SEARCH, BMTSyncing_FinishNarrowingSearch, realmName)
end

function BMTSyncing_BroadcastAuctionsRequest()
	if BMTSyncingState ~= SYNCING_STATE_NONE then return end
	
	local realmName = GetRealmName()
	BMT_RefreshAllPredictedTimes(realmName)
	local msg = BMTSyncing_InitialInfo(KEY_BETTER_SEARCH_REQ)
	-- if we have nothing to send, send ITEMS_NONE
	if BMTGlobal.current[realmName].updateInfo.playerName == nil then
		msg = msg .. ITEMS_NONE
	else
		-- last one has different separator
		for i=1, #updateInfoSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BMTGlobal.current[realmName].updateInfo[updateInfoSyncField[i]]
		end
		msg = msg .. SEPARATOR_CATEGORY
		-- send out short details about each item, so that players can determine wheather some of their items are outdated or not, as well as determine new time predictions for each
		local numItems = BMT_GetNumItems(realmName)
		for j=1, numItems do
			for i=1, #itemShortSyncField do
				if i > 1 then
					msg = msg .. SEPARATOR
				end
				msg = msg..BMTGlobal.current[realmName].items[j][itemShortSyncField[i]]
			end
			if j < numItems then
				msg = msg .. SEPARATOR_CATEGORY
			end
		end
	end
	BMTSyncing_BroadcastAddonMessage(msg)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_BETTER_SEARCH_REQ, BMTSyncing_FinishNarrowingSearch, realmName)
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_BETTER_SEARCH)
end

function BMTSyncing_BroadcastFullItemDatabase(realmName)
	if BMTGlobal.current == nil or BMTGlobal.current[realmName] == nil or BMTGlobal.current[realmName].updateInfo == nil or BMTGlobal.current[realmName].updateInfo.playerName == nil or #BMTGlobal.current[realmName].items == 0 then return end
	BMT_RefreshAllPredictedTimes(realmName)
	local msg = BMTSyncing_InitialInfo(KEY_ITEM_FULL)
	-- update info
	for i=1, #updateInfoSyncField do
		if i > 1 then
			msg = msg .. SEPARATOR
		end
		msg = msg..BMTGlobal.current[realmName].updateInfo[updateInfoSyncField[i]]
	end
	msg = msg .. SEPARATOR_CATEGORY
	-- full item detauls
	local numItems = BMT_GetNumItems(realmName)
	for j=1, numItems do
		for i=1, #itemFullSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BMTGlobal.current[realmName].items[j][itemFullSyncField[i]]
		end
		if j < numItems then
			msg = msg .. SEPARATOR_CATEGORY
		end
	end
	BMTSyncing_BroadcastAddonMessage(msg)
end

---- TIMEOUTS ----
-- bottom text handling, also asks for auctions
function BMTSyncing_FinishSyncingVersion()
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
	-- fade out only if we're not fading already
	if (BMTHighestRevision > BMTRevision and BMTFadeState == 3) and BMTFrame:IsShown() then
		BMTFadeState = -1
		BMTBottomText_FadingHandler()
	end
	BMTSyncing_BroadcastAuctionsRequest()
end

-- this function is called when timeout finishes for both: new search sender, login request sender
function BMTSyncing_FinishNarrowingSearch(syncedRealm)
	BMT:Debug("BMTSyncing_FinishNarrowingSearch(), BMTSyncingState="..BMTSyncingState.." syncedRealm="..syncedRealm)
	-- this has done new search and might've received better predictions, check and set them here
	if BMTSyncingState == SYNCING_STATE_NARROWING_SEARCH and BMTSyncBetterDB ~= nil and BMTSyncBetterDB[syncedRealm] ~= nil then
		-- here we will ignore the update infos completly, as we're just after doing a search, therefore our should be the newest one
		local numBetterDB = #BMTSyncBetterDB[syncedRealm]
		local numItems = #BMTGlobal.current[syncedRealm].items
		-- won't run if received nothing better
		-- go through all received better results
		for k=1, numBetterDB do
			-- go through all of our items
			for i=1, numItems do
				local numRecvItems = #BMTSyncBetterDB[syncedRealm][k].items
				-- go through all received items from one player
				for j=1, numRecvItems do
					if BMTSyncBetterDB[syncedRealm][k].items[j].marketID == BMTGlobal.current[syncedRealm].items[i].marketID then
						local foundMarketID = BMTGlobal.current[syncedRealm].items[i].marketID
						-- set best to our current
						local bestPredictedLow = BMTGlobal.current[syncedRealm].items[i].predictedLow
						local bestPredictedHigh = BMTGlobal.current[syncedRealm].items[i].predictedHigh
						-- go through received databases from other players, trying to find better predictions
						for l=1, numBetterDB do
							-- go through all items for each player, if marketID matches - check if prediction is better
							local numRecvPlrItems = #BMTSyncBetterDB[syncedRealm][l].items
							for m=1, numRecvPlrItems do
								if foundMarketID == BMTSyncBetterDB[syncedRealm][l].items[m].marketID then
									if BMTSyncBetterDB[syncedRealm][l].items[m].predictedLow > bestPredictedLow then
										bestPredictedLow = BMTSyncBetterDB[syncedRealm][l].items[m].predictedLow
									end
									if BMTSyncBetterDB[syncedRealm][l].items[m].predictedHigh < bestPredictedLow then
										bestPredictedHigh = BMTSyncBetterDB[syncedRealm][l].items[m].predictedHigh
									end
									break
								end
							end
						end
						-- here, we should have best predictions for one item, check for weird stuff
						if bestPredictedLow > bestPredictedHigh then break
						else
							BMTGlobal.current[syncedRealm].items[i].predictedLow = bestPredictedLow
							BMTGlobal.current[syncedRealm].items[i].predictedHigh = bestPredictedHigh
						end
						break
					end
				end
			end
		end
	-- logged in and asked for newer searches/better responses - maybe got them
	-- if received nothing then don't do anything here
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH and BMTSyncBetterDB ~= nil and BMTSyncBetterDB[syncedRealm] ~= nil then
		BMT:Debug("BMTSyncing_FinishNarrowingSearch() @ SYNCING_STATE_AWAITING_BETTER_SEARCH")
		-- newest search isn't necessarily the most accurate on predictions
		-- algorithm is:
		-- a. recalculate predictions to current time (locally only) - because browser could be opened after logging in but before this function got called
		-- b. move all predictions back up again, to the point of our last search
		-- c. get newest search update info
		-- d. remove old auctions, accordingly to items in the newest search
		-- e. recalculate the predictions, as if the search was done at the point of newest search ('fake search')
		-- f. recalculate predictions to current time (locally only)
		
		
		-- true if this has empty DB
		local emptyDB = false
		if BMTGlobal.current[syncedRealm] == nil or BMTGlobal.current[syncedRealm].updateInfo == nil or BMTGlobal.current[syncedRealm].updateInfo.playerName  == nil or #BMTGlobal.current[syncedRealm].items == 0 then
			emptyDB = true
		end
		
		if emptyDB then
			BMT:Debug("emptyDB = true")
		end
		
		-- a. recalculate predictions to current time (locally only) - because browser could be opened after logging in but before this function got called
		BMT_RefreshAllPredictedTimes(syncedRealm)
		
		-- b. move all predictions back up again, to the point of our last search
		BMT_MoveAllPredictionsToLastSearchTime(syncedRealm)
		
		-- c. get newest search update info (nil is current)
		local newestUpdateInfoIndex = nil
		local numBetterDB = #BMTSyncBetterDB[syncedRealm]
		BMT:Debug("numBetterDB="..numBetterDB)
		for k=1, numBetterDB do
			if newestUpdateInfoIndex == nil then
				-- our is older
				if not emptyDB then
					if BMT_CompareDateAndTime(BMTGlobal.current[syncedRealm].updateInfo.updateDate, BMTGlobal.current[syncedRealm].updateInfo.updateTime, BMTSyncBetterDB[syncedRealm][k].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][k].updateInfo.updateTime) == "smaller" then
						newestUpdateInfoIndex = k
					end
				else
					newestUpdateInfoIndex = k
				end
			else
				-- previous best is older
				if BMT_CompareDateAndTime(BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateTime, BMTSyncBetterDB[syncedRealm][k].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][k].updateInfo.updateTime) == "smaller" then
					newestUpdateInfoIndex = k
				end
			end
		end
		-- if nothing better was found, we'll just send current, after moving predictions back to current 
		-- newer search was found
		if newestUpdateInfoIndex ~= nil and not emptyDB then
			BMT:Debug("newestUpdateInfoIndex="..newestUpdateInfoIndex)
			-- d. remove old auctions, accordingly to items in the newest search
			BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].items, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateTime)
			
			-- e. recalculate the predictions, as if the search was done at the point of newest search ('fake search')
			local numItems = BMT_GetNumItems(syncedRealm)
			for i=1, numItems do
				local newTimeLeft = BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].items[i].timeLeft
				BMT_RecalculateAndSetPredictedTimes(syncedRealm, i, newTimeLeft, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateDate, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].updateInfo.updateTime, false)
			end
			-- if DB was empty - rewrite everything
			if emptyDB then
				BMTGlobal.current[syncedRealm].items = BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex].items
			end
		end
		-- f. recalculate predictions to current time (locally only)
		BMT_RefreshAllPredictedTimes(syncedRealm)
	end
	BMT:Debug("Broadcasting db")
	BMTSyncing_BroadcastFullItemDatabase(syncedRealm)
	BMTSyncBetterDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
end


---- SPINNING TILES ----
function BMTSpinningInset_Initialize()
	-- if you change it here, change it in the BlackMarketTracker.xml
	local self = BMTFrameSpinningInset
	-- set rotations
	for i=0, 7 do
		_G[self:GetName().."Tile"..i]:SetRotation(BMTDegToRad(i*45));
		-- original size is 32x32, which is a little to big in this case
		_G[self:GetName().."Tile"..i]:SetWidth(SPINNING_TILE_SIZE);
		_G[self:GetName().."Tile"..i]:SetHeight(SPINNING_TILE_SIZE);
	end
	-- hide all tiles
	for i=0, 7 do
		_G[self:GetName().."Tile"..i]:SetAlpha(0);
	end
	-- fade in first tile
	BMTTileFadeIn(_G[self:GetName().."Tile0"], 0);
end

function BMTDegToRad(degrees)
	return ((PI/180)*degrees);
end

function BMTFadeInTile(frame, id)
	local fadeInfo={
    mode = "IN",
    timeToFade = SPINNING_TILE_FADE_IN_TIME,
    finishedFunc = BMTTileFadeIn,
    finishedArg1 = frame,
	finishedArg2 = id,
    }
    UIFrameFade(frame, fadeInfo);
end

function BMTTileFadeIn(self, id)
	local fadeInfo={
    mode = "OUT",
    timeToFade = SPINNING_TILE_FADE_OUT_TIME,
    finishedFunc = BMTTileFadeOut,
    finishedArg1 = self,
	finishedArg2 = id,
    }
    UIFrameFade(self, fadeInfo);

	-- fade in next too
	local nextId, nextTile;
	if (id>0) then
		nextId = id-1;
	else
		nextId = 7;
	end
	
	nextTile = _G[self:GetParent():GetName().."Tile"..nextId];
	BMTFadeInTile(nextTile, nextId);
end

-- for future use, if ever needed
function BMTTileFadeOut(self, id)	
end
